/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.*;
import org.wso2.registry.i18n.Messages;
import org.wso2.registry.jdbc.dao.CommentsDAO;
import org.wso2.registry.jdbc.dao.RatingsDAO;
import org.wso2.registry.jdbc.dao.TagsDAO;
import org.wso2.registry.jdbc.dao.VersionedArtifactDAO;
import org.wso2.registry.jdbc.dataobjects.TagDO;
import org.wso2.registry.jdbc.mediatypes.MediaTypeManager;
import org.wso2.registry.jdbc.queries.QueryProcessor;
import org.wso2.registry.jdbc.queries.QueryProcessorFactory;
import org.wso2.registry.jdbc.urlhandlers.URLHandlerManager;
import org.wso2.usermanager.Realm;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Date;

/**
 * JDBC based implementation of the Registry. This will be used mostly as the back-end by other
 * Registry implementations. This can use either an in-memory database or a external database
 * configured using data source. At the initializing, it checks if a data source named
 * "jdbc/WSO2RegistryDB" is configured. If such data source is found, it is used as the database.
 * If a data source is not found, an in-memory database is created using the HSQL.
 */
public class JDBCRegistry implements Registry {

    private static final Log log = LogFactory.getLog(JDBCRegistry.class);

    private ConnectionFactory connectionFactory = null;
    private URLHandlerManager urlHandlerManager;
    private MediaTypeManager mediaTypeManager;
    private VersionedArtifactDAO artifactDAO = null;
    private TagsDAO tagsDAO = null;
    private CommentsDAO commentsDAO = null;
    private RatingsDAO ratingsDAO = null;

    public JDBCRegistry(Realm realm) throws RegistryException {
        connectionFactory = new ConnectionFactory();
        urlHandlerManager = new URLHandlerManager(connectionFactory);
        mediaTypeManager = new MediaTypeManager(connectionFactory, realm);

        this.artifactDAO = new VersionedArtifactDAO();
        this.tagsDAO = new TagsDAO();
        this.commentsDAO = new CommentsDAO();
        this.ratingsDAO = new RatingsDAO();

        // check if root is added. if not add it.

        Connection conn = connectionFactory.getConnection();

        try {
            conn.setAutoCommit(false);

            Artifact root = artifactDAO.getLatestVersion(RegistryConstants.ROOT_PATH, conn);
            if (root == null) {
                root = new Artifact();
                root.setPath(RegistryConstants.ROOT_PATH);
                root.setDirectory(true);
                artifactDAO.add(RegistryConstants.ROOT_PATH, root, conn);
                //authorizationUtil.setDefaultAuthorizations(realm, RegistryConstants.ROOT_PATH, );

                root = artifactDAO.getLatestVersion(RegistryConstants.ROOT_PATH, conn);
                artifactDAO.addResourceVersion(root, conn);
            }

            conn.commit();

        } catch (SQLException e) {

            try {
                conn.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }

            String msg = Messages.getMessage("jdbc.reg.init.fail");
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {

            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Returns the resource in the given path.
     *
     * @param path Path of the resource. e.g. /project1/server/deployment.xml
     * @return Resource instance
     * @throws RegistryException ResourceNotFound exception is thrown if the resource is not in the
     *                           registry
     */
    public Artifact get(String path) throws RegistryException {

        path = preparePath(path);
        Artifact artifact;

        // check if this path referes to a resource referred by a URL query (e.g. comment)
        artifact = urlHandlerManager.handleURL(path);

        // check if the resource in the path has a mediatype handler
        if (artifact == null) {
            artifact = mediaTypeManager.get(path);
        }

        if (artifact == null) {
            String message = Messages.getMessage("artifact.get.error", path);
            throw new RegistryException(message);
        }

        return artifact;
    }

    public boolean resourceExists(String path) throws RegistryException {

        Connection conn = connectionFactory.getConnection();

        boolean resourceExist = false;

        try {
            resourceExist = artifactDAO.artifactExist(path, conn);

        } catch (SQLException e) {
            String msg = "Failed to check the exsitence of a resource at path " + path;
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

        return resourceExist;
    }

    /**
     * Adds or updates resources in the registry. If there is no resource in the given path,
     * resource is added. If a resource already exist in the given path, it will be replaced with
     * the new resource. Implementations may use a concurrency control method to prevent
     * concurrency issues.
     *
     * @param path   Path of the new resource
     * @param artifact Resource instance for the new resource
     * @throws org.wso2.registry.RegistryException
     *          : if the user is not authorized
     */
    public synchronized void put(String path, Artifact artifact) throws RegistryException {
        path = preparePath(path);
        mediaTypeManager.put(path, artifact);
    }

    /**
     * Deletes the resource in the given path. If the path refers to a directory, all child
     * resources of the directory will also be deleted.
     *
     * @param path Path of the resource to be deleted.
     * @throws org.wso2.registry.RegistryException
     *          : if the user is not authorized
     */
    public synchronized void delete(String path) throws RegistryException {

        path = preparePath(path);

        mediaTypeManager.delete(path);        
    }

    public String[] getVersions(String path) throws RegistryException {

        String[] versionPaths = null;

        Connection conn = connectionFactory.getConnection();

        try {
            long artifactID = artifactDAO.getArtifactID(path, conn);
            String[] versionIDs = artifactDAO.getVersionNumbers(artifactID, conn);

            versionPaths = new String[versionIDs.length];
            for (int i = 0; i < versionIDs.length; i++) {
                versionPaths[i] = path + "?v=" + versionIDs[i];
            }

        } catch (SQLException e) {

            try {
                conn.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }

            String msg = "Could not get the version paths of the resource " +
                    path + ". Caused by: " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException ignore) {
                log.info("Nothing we can do here : Inside the finally block ");
            }
        }

        return versionPaths;
    }

    public void restoreVersion(String versionPath) throws RegistryException {
        String plainPath ;
        long versionNumber = -1;
        if (versionPath.indexOf("?v=") != -1) {
            String[] parts = versionPath.split("\\?v=");
            plainPath = parts[0];
            versionNumber = new Long(parts[1]).longValue();
        } else {
            String msg = "Given path is not a path of a versioned resource.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn = connectionFactory.getConnection();

        try {
            conn.setAutoCommit(false);

            long artifactID = artifactDAO.getArtifactID(plainPath, conn);

            artifactDAO.restore(artifactID, versionNumber, conn);
            //artifactDAO.restore(artifactID, versionNumber, conn);
            //VersionHandler.revert(artifactID, versionNumber, conn);

            conn.commit();

        } catch (SQLException e) {

            String msg = "Could not restore the version " +
                    versionPath + ". Caused by: " + e.getMessage();
            log.error(msg);

            try {
                conn.rollback();
            } catch (SQLException e1) {
                log.error(e1);
            }

            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException ignore) {}
        }
    }

    ////////////////////////////////////////////////////////
    // Tagging
    ////////////////////////////////////////////////////////

    /**
     * Applies the given tag to the resource in the given path. If the given tag is not defined
     * in the registry, it will be defined.
     *
     * @param resourcePath Path of the resource to be tagged.
     * @param tag          Tag. Any string can be used for the tag.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void applyTag(String resourcePath, String tag) throws RegistryException {

        resourcePath = preparePath(resourcePath);

        Connection conn = connectionFactory.getConnection();

        String userID = User.getCurrentUser();

        // break the space separated words into multiple tags
        String[] tags = tag.split(" ");

        try {
            conn.setAutoCommit(false);

            if (!artifactDAO.artifactExist(resourcePath, conn)) {
                throw new RegistryException(
                        "Cannot tag on a non-existing resource " + resourcePath);
            }
            Artifact artifact = artifactDAO.getLatestVersion(resourcePath, conn);

            for (int i = 0; i < tags.length; i++) {

                if (tags[i].length() == 0 || tags[i].equals(" ")) {
                    continue;
                }

                if (!tagsDAO.tagExist(tags[i], conn)) {
                    tagsDAO.addTag(tags[i], conn);
                }

                TagDO tagDO = tagsDAO.getTag(tags[i], conn);

                if (tagsDAO.taggingExist(tagDO.getId(), artifact.getId(), userID, conn)) {
                    String msg =
                            Messages.getMessage("resource.already.tagged", resourcePath, tags[i], userID);
                    log.error(msg);
                    throw new RegistryException(msg);
                }

                tagsDAO.addTagging(tagDO.getId(), artifact.getId(), userID, conn);
            }

            conn.commit();

        } catch (SQLException e) {
            try {
                conn.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }

            String msg = Messages.getMessage("resource.tagging.fail", resourcePath, tag);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Returns the paths of the resources, which are tagged using the given tag.
     *
     * @param tag Tag
     * @return Resource paths
     */
    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        // break the tags from spaces
        String[] tags = tag.split(" ");

        List pathList = new ArrayList();

        Connection conn = connectionFactory.getConnection();
        try {
            for (int i = 0; i < tags.length; i++) {
                if (tags[i].length() == 0 || tags[i].equals(" ")) {
                    continue;
                }

                pathList.addAll(tagsDAO.getTaggedPaths(tags[i], conn));
            }

        } catch (SQLException e) {

            String msg = Messages.getMessage("tag.search.fail.1", tag);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

        return
                (TaggedResourcePath[]) pathList.toArray(new TaggedResourcePath[pathList.size()]);
    }

    /**
     * Returns all tags used for tagging the given resource.
     *
     * @param resourcePath Path of the resource
     * @return Tags
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public Tag[] getTags(String resourcePath) throws RegistryException {

        resourcePath = preparePath(resourcePath);

        Connection conn = connectionFactory.getConnection();

        Tag[] tags = null;

        try {
            tags = tagsDAO.getTagsWithCount(resourcePath, conn);

        } catch (SQLException e) {

            String msg = Messages.getMessage("tag.search.fail.2", resourcePath);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

        return tags;
    }

    ////////////////////////////////////////////////////////
    // Comments
    ////////////////////////////////////////////////////////

    /**
     * Adds a comment to a resource.
     *
     * @param resourcePath Path of the resource to add the comment.
     * @param comment      Comment instance for the new comment.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void addComment(String resourcePath, Comment comment)
            throws RegistryException {

        resourcePath = preparePath(resourcePath);

        Connection conn = connectionFactory.getConnection();

        String userID = User.getCurrentUser();

        try {

            conn.setAutoCommit(false);

            Artifact artifact = artifactDAO.getLatestVersion(resourcePath, conn);
            if (artifact != null) {
                commentsDAO.addComment(artifact.getId(), userID, comment, conn);
            } else {

                String msg = Messages.getMessage("comment.on.null.artfact", resourcePath);
                log.error(msg);
                throw new RegistryException(msg);
            }

            conn.commit();

        } catch (SQLException e) {

            try {
                conn.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }

            String msg = Messages.
                    getMessage("resource.comment.fail", resourcePath, comment.getCommentText());
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {

            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

    }

    /**
     * Retruns all comments for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Array of Comment objects.
     */
    public Comment[] getComments(String resourcePath) throws RegistryException {

        resourcePath = preparePath(resourcePath);

        Connection conn = connectionFactory.getConnection();

        Comment[] comments = new Comment[0];

        try {
            comments = commentsDAO.getComments(resourcePath, conn);
        } catch (SQLException e) {

            String msg = Messages.getMessage("get.comments.fail", resourcePath);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

        return comments;
    }

    ////////////////////////////////////////////////////////
    // Ratings
    ////////////////////////////////////////////////////////

    /**
     * Rates the given resource.
     *
     * @param resourcePath Path of the resource.
     * @param rating       rating value.
     * @throws RegistryException ResourceNotFound exception is thrown if a resource does not exist
     *                           in the given path.
     */
    public synchronized void rateResource(String resourcePath, int rating) throws RegistryException {

        resourcePath = preparePath(resourcePath);

        Connection conn = connectionFactory.getConnection();

        String userID = User.getCurrentUser();

        try {
            conn.setAutoCommit(false);

            Artifact artifact = artifactDAO.getLatestVersion(resourcePath, conn);

            if (artifact != null) {
                if (ratingsDAO.ratingExist(artifact.getId(), userID, conn)) {
                    ratingsDAO.updateRating(artifact.getId(), userID, rating, conn);
                    log.info("Updated the rating on the resource " +
                            resourcePath + " by user " + userID);
                } else {
                    ratingsDAO.addRating(artifact.getId(), userID, rating, conn);
                }
            } else {
                String msg = Messages.getMessage("rate.on.null.artfact", resourcePath);
                log.error(msg);
                throw new RegistryException(msg);
            }

            conn.commit();

        } catch (SQLException e) {

            try {
                conn.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }

            String msg = Messages.getMessage("resource.rating.fail", resourcePath);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {

            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }


    }

    /**
     * Returns the avarage rating for the given resource. This is the average of all ratings done by
     * all users for the given resource.
     *
     * @param resourcePath Path of the resource.
     * @return Average rating.
     */
    public int getAverageRating(String resourcePath) throws RegistryException {

        resourcePath = preparePath(resourcePath);

        Connection conn = connectionFactory.getConnection();

        int rating = 0;

        try {
            rating = ratingsDAO.getAverageRating(resourcePath, conn);
        } catch (SQLException e) {

            String msg = Messages.getMessage("get.average.rating.fail", resourcePath);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

        return rating;
    }

    public int getRating(String path, String userName) throws RegistryException {

        path = preparePath(path);

        Connection conn = connectionFactory.getConnection();

        int rating = 0;

        try {
            rating = ratingsDAO.getRating(path, userName, conn);
        } catch (SQLException e) {

            String msg = "Could not get the rating of the resource " + path +
                    " given by the user " + userName + ". Caused by: " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }

        return rating;
    }

    //////////////////////////////////////////////////////////////////////////////
    // Extensible searching API
    //////////////////////////////////////////////////////////////////////////////

    public synchronized void defineQuery (String name, String query) throws RegistryException {

        Connection conn = connectionFactory.getConnection();

        String userID = User.getCurrentUser();

        // make the path of the query (e.g. /users/chathura/queries/myquery1 )
        String queryPath = RegistryConstants.ROOT_PATH + RegistryConstants.USERS_COLLECTION_NAME +
                RegistryConstants.PATH_SEPARATOR + userID + RegistryConstants.PATH_SEPARATOR +
                RegistryConstants.QUERIES_COLLECTION_NAME + RegistryConstants.PATH_SEPARATOR + name;


        try {

            conn.setAutoCommit(false);

            Artifact queryArtifact = new Artifact();
            queryArtifact.setPath(queryPath);
            queryArtifact.setMediaType(RegistryConstants.SQL_QUERY_MEDIA_TYPE);
            queryArtifact.setAuthorUserName(userID);
            queryArtifact.setContent(query.getBytes());

            Artifact artifact = artifactDAO.getLatestVersion(queryPath, conn);
            if (artifact != null) {
                artifactDAO.update(queryPath, queryArtifact, conn);
                log.info("Updated the predefined query: " + queryPath);
            } else {
                artifactDAO.add(queryPath, queryArtifact, conn);
            }

            conn.commit();

        } catch (SQLException e) {

            try {
                conn.rollback();
            } catch (SQLException e1) {
                e1.printStackTrace();
            }

            String msg = Messages.getMessage("define.query.fail", name, query);
            log.error(msg, e);
            throw new RegistryException(msg);

        } finally {
            try {
                conn.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }


    }

    public Artifact executeQuery (String path, Object[] parameters) throws RegistryException {
        Connection conn = connectionFactory.getConnection();
        try {
            Artifact query = get(path);
            QueryProcessor queryProcessor =
                    new QueryProcessorFactory().getQueryProcessor(query.getMediaType());
            if (queryProcessor == null) {
                throw new RegistryException("Undefined query type: " + query.getMediaType());
            }
            return queryProcessor.executeQuery(query, parameters, conn);
        } finally {
            try {
                conn.close();
            } catch (SQLException ignore) {}
        }
    }
    public void shutdown() throws RegistryException {
        try {
            connectionFactory.shutdown();
        } catch (Exception e) {
            throw new RegistryException("Unable to shutdown the database.");
        }
    }
    private String preparePath(String rawPath) {

        String path = rawPath;

        // path should always start with "/"
        if (!rawPath.startsWith(RegistryConstants.ROOT_PATH)) {
            path = RegistryConstants.ROOT_PATH + rawPath;
        }

        return path;
    }

    public LogEntry[] getLogs(String resourcePath, String userName, Date from, Date to) throws RegistryException {
        return new LogEntry[0];  //To change body of implemented methods use File | Settings | File Templates.
    }
}
